from collections.abc import Sequence
from itertools import cycle
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import seaborn as sns
from IPython.display import display

from . import dist_distrib as dd
from . import fret


def mpl_text_1state(params, E_μ, show_τ_relax=True, space='nm', time='s'):
    text = rf'R ~ OU{{μ={params["R_mean"][0]:.1f} {space}, σ={params["R_sigma"][0]:.1f} {space}'
    if show_τ_relax:
        text += ',\n' + rf'$\qquad\qquad\;\tau_{{relax}}$={params["τ_relax"][0]:,} {time}'
    text += ('}\n'
             rf'$R_0$ = {params["R0"]} {space},  $\langle E \rangle$ = {E_μ:.2f}' + '\n'
             rf'$\tau_D$ = {params["τ_D"]} {time},  $\tau_A$ = {params["τ_A"]} {time}' + '\n'
             rf'$\delta t$ = {params["δt"]:,} {time}' + '\n'
             )
    return text


def plot_text_1state(ax, x, y, params, show_τ_relax=True, **text_kws):
    p = params
    E_μ = fret.mean_E_from_gauss_PoR(p['R_mean'][0], p['R_sigma'][0], p['R0'])
    text_kws_used = dict(transform=ax.transAxes, va='top')
    text_kws_used.update(**text_kws)
    text = mpl_text_1state(params, E_μ, show_τ_relax=show_τ_relax)
    ax.text(x, y, text, **text_kws_used)


def plot_R_distrib(params, ax=None, dr=0.01, show_html=True):
    if ax is None:
        _, ax = plt.subplots()
    d = dd.distribution(params)
    x, y = d.get_pdf(dr=dr)
    if show_html:
        display(d)
    ax.plot(x, y, '-k')
    ax.axvline(params['R0'], color='r', ls='--')
    return ax


def plot_E_sim(burstsph_sim, params, E=None, ax=None, legend=True):
    if ax is None:
        _, ax = plt.subplots()
    if E is None:
        E = fret.calc_E_burst(burstsph_sim)
    bins = np.arange(-0.1, 1.1, 0.02)
    R0 = params['R0']
    E_ph = fret.E_from_dist(burstsph_sim.R_ph, R0)
    E_burst = E_ph.groupby('burst').mean()
    ax.hist([], histtype='step', lw=1.2, color='k',
            label='E per ph (no bursts, y-axis A.U.)')
    ax.hist(E, bins=bins, alpha=0.8, label='E per burst (w/ shot-noise)')
    ax2 = ax.twinx()
    ax2.hist(E_ph, histtype='step', weights=np.repeat(0.5e-1, E_ph.size),
             lw=1.2, color='k', bins=100)
    ax.hist(E_burst, bins=bins, alpha=0.6, color='C1',
            label='Burst avg. of E from R @ t_emission')
    if params['name'].lower().startswith('gauss'):
        R_mean = np.atleast_1d(params['R_mean'])
        R_sigma = np.atleast_1d(params['R_sigma'])
        for i, (R_m, R_s) in enumerate(zip(R_mean, R_sigma)):
            E_μ = fret.mean_E_from_gauss_PoR(R_m, R_s, R0)
            ax.axvline(E_μ, ls='--', color='k', label=f'state{i}')
    #ax.set_xlabel('E')
    if legend:
        ax.legend(loc=(0.9, 0.0), frameon=False, ncol=1)
        #ax2.legend(loc=(1, 0), frameon=False);
    sns.despine(ax=ax, trim=True)
    sns.despine(ax=ax2, left=True, trim=True)
    ax2.yaxis.set_visible(False)
    return ax


def plot_nanotimes(burstsph, ax=None):
    if ax is None:
        _, ax = plt.subplots()
    nt_d = burstsph.loc[burstsph.stream == 'DexDem', 'nanotime']
    nt_a = burstsph.loc[burstsph.stream == 'DexAem', 'nanotime']
    kws = dict(bins=np.arange(0, 50, 0.2), log=True, histtype='step', lw=1.3)
    ax.hist(nt_d, color='C2', label='D', **kws)
    ax.hist(nt_a, color='C3', label='A', **kws)
    sns.despine(ax=ax)
    return ax


def plot_nanotime_tau_rel_series(relax_dict, cmap='Spectral', axes=None):
    colors = sns.color_palette(cmap, len(relax_dict))
    if axes is None:
        _, axes = plt.subplots(2, 1, figsize=(8, 8))
    for color, (τ_relax, p) in zip(cycle(colors), relax_dict.items()):
        burstsph = p['bph']
        nt_d = burstsph.loc[burstsph.stream == 'DexDem', 'nanotime']
        nt_a = burstsph.loc[burstsph.stream == 'DexAem', 'nanotime']
        kws = dict(label=f'{τ_relax:,} ns', bins=np.arange(0, 50, 0.2),
                   log=True, histtype='step', lw=1.3, color=color)
        axes[0].hist(nt_d, **kws)
        axes[1].hist(nt_a, **kws)
    for a, label in zip(axes, ['Donor', 'Acceptor']):
        sns.despine(ax=a)
        a.set_xlabel('ns')
        a.set_title(label, y=0.9, va='top')
        a.set_ylim(0.5)
    return axes, colors


def plot_E_tau_rel_series(relax_dict, cmap=None, axes=None, colors=None,
                          legend=True, **kws):
    if colors is None:
        if cmap is not None:
            colors = sns.color_palette(cmap, len(relax_dict))
        else:
            colors = [f'C{i}' for i in range(10)]
    if axes is None:
        _, axes = plt.subplots(1, 3, figsize=(10, 4), sharex=True)
    bins = np.arange(-0.1, 1.1, 0.0333)
    for color, (τ_relax, p) in zip(cycle(colors), relax_dict.items()):
        bph = p['bph']
        params = p['params']
        R0 = params['R0']
        E = fret.calc_E_burst(bph)
        E_μ = fret.mean_E_from_gauss_PoR(params['R_mean'][0], params['R_sigma'][0], R0)
        #E_mode = fret.E_from_dist(params['R_mean'][0], R0)
        E_ph = fret.E_from_dist(bph.R_ph, R0)
        #E_burst = E_ph.groupby('burst').mean()
        hist_kws = dict(histtype='stepfilled', color=color,
                        label=f'{τ_relax:,} ns', bins=bins)
        hist_kws.update(kws)
        axes[0].hist(E_ph, **hist_kws)
        axes[1].hist(E_ph.groupby('burst').mean(), **hist_kws)
        axes[2].hist(E, **hist_kws)
    labels = ['E from R, per ph', 'E from R, burst average', 'E burst (incl. shot noise)']
    for a, t in zip(axes, labels):
        sns.despine(ax=a)
        a.set_title(t)
        a.axvline(E_μ, color='k', ls='-', alpha=0.3)
        #a.axvline(E_mode, color='k', ls=':')
    if legend:
        axes[-1].legend(bbox_to_anchor=(1.1, 1), loc='upper right',
                        title=r'$\tau_{relax}$', frameon=True)
    return axes, colors


def plot_colorbar(relax_dict, colors, vertical=False, ax=None):
    if all(c in [f'C{i}' for i in range(10)] for c in colors):
        #idx_c  = [int(c[1]) for c in colors]
        hex_c = plt.rcParams['axes.prop_cycle'].by_key()['color']
        colors = [hex_c[i] for i in idx_i]
    if isinstance(colors[0], str):
        colors = [mpl.colors.hex2color(c) for c in colors]
    if vertical:
        if ax is None:
            ax = plt.axes([1, 0.25, 0.02, 0.5])
        ax.yaxis.tick_right()
        ax.imshow(np.asarray(colors)[:, np.newaxis, :], aspect=3)
        ax.xaxis.set_visible(False)
        ax.set_yticks(range(len(relax_dict)))
        ax.set_yticklabels(['%g' % r for r in relax_dict])
    else:
        if ax is None:
            ax = plt.axes([0.25, 0, 0.5, 0.02])
        ax.imshow(np.asarray(colors)[np.newaxis, ...], aspect=0.3)
        ax.yaxis.set_visible(False)
        ax.set_xticks(range(len(relax_dict)))
        ax.set_xticklabels(['%g' % r for r in relax_dict])
    ax.set_title(r'$\tau_{relax}$ (ns)')
    return ax


def plot_E_R_pdf(params, ax=None, which='ER', **kws):
    assert len(which) <= 2, '`which` can only be "E", "R" or "ER"'
    R0 = params['R0']
    R_mean = params['R_mean']
    R_sigma = params['R_sigma']
    R_samples = np.random.randn(500000) * R_sigma + R_mean
    E_samples = fret.E_from_dist(R_samples, R0)
    style = dict(color='k', lw=2, histtype='step', density=True)
    style.update(kws)
    if ax is None:
        _, ax = plt.subplots(1, len(which), figsize=(6 * len(which), 4))
    if isinstance(ax, Sequence) or hasattr(ax, '__array__'):
        axR, axE = ax
        which = 'ER'
    else:
        assert len(which) == 1, 'One axis passed but two chars in `which`'
        if which == 'E':
            axE = ax
        elif which == 'R':
            axR = ax
    if 'R' in which:
        axR.hist(R_samples, bins=200, **style)
    if 'E' in which:
        axE.hist(E_samples, bins=np.arange(0, 1.01, 0.01), **style)
        axE.set_ylim(0)
    return ax



# To build the cython extensions use:
# python setup.py build_ext --inplace
#

from distutils.core import setup
from distutils.extension import Extension
from Cython.Build import cythonize
import numpy as np

extensions = [
    Extension(
        'depi_cy',
        ['depi_cy.pyx'],
        include_dirs=[np.get_include()],
        # Uncomment the following line for profiling
        #define_macros=[('CYTHON_TRACE', '1')],
    )
]

setup(
    name="DEPI with intra-lifetime diffusion core functions",
    ext_modules=cythonize(
        extensions,
        # Uncomment the following line for profiling
        #compiler_directives={'linetrace': True, 'binding': True},
    ),
)



"""
Utility functions to work with TCSPC nanotimes decays.

License: MIT
Copyright 2017 Antonino Ingargiola <tritemio@gmail.com>
"""

import numpy as np


def fullwidth_alphamax(x, y_peak, alpha):
    """Compute full-width at fraction `alpha` of max of passed curve.

    Computes the curve full-width at given percentage of the max.
    To compute the FWHM use `alpha = 0.5`.

    Arguments:
        x, y_peak (arrays): the peak coordinates. It must have a monotonic
            rise and a monotonic fall.
        alpha (float): fraction of the peak max at which the full width
            is computed. Valid values are in the range (0, 1).

    Returns:
        Tuple of two floats: full-width of the peak and x-axis position where
        the raising edge reaches a fraction `alpha` of its max.
    """
    y_unit = y_peak / y_peak.max()
    t_rise = np.interp(alpha,
                       xp=y_unit[:y_unit.argmax()],
                       fp=x[:y_unit.argmax()])
    idx_fall_stop = np.where(y_unit[y_unit.argmax():] < alpha)[0][0]
    xp = y_unit[y_unit.argmax():y_unit.argmax() + idx_fall_stop + 1][::-1]
    fp = x[y_unit.argmax():y_unit.argmax() + idx_fall_stop + 1][::-1]
    assert (np.diff(xp) >= 0).all()
    t_fall = np.interp(alpha, xp=xp, fp=fp)
    fullwidth = t_fall - t_rise
    return fullwidth, t_rise


def calc_nanotime_hist(nanotimes, unit, idxmax, idxmin=0, rebin=1,
                       t_start=-np.inf, t_stop=np.inf, offset=0, pdf=False):
    """Compute nanotimes histogram.

    Arguments `t_start`, `t_stop`, `offset` and `unit` need to use in the
    same "units", for example nanoseconds.

    Arguments:
        idxmin, idxmax (int): min and max bin range in raw nanotimes units.
        rebin (int): bin width in raw nanotimes units.
        offset (float): time (in `unit`) to use as time=0 in the histogram
            time axis.
        pdf (bool): if True normalize the histogram to be a PDF.
        t_start, t_stop (float): min/max time range (in `unit`) for the
            returned histogram. The range selection is applied after
            translating nanotimes by `offset`.

    Returns:
        Two arrays: bin centers in `units` and histogram of nanotimes.
    """
    bins_raw = np.arange(idxmin, idxmax, rebin)

    nanot_hist, _ = np.histogram(nanotimes, bins=bins_raw)
    bin_centers = ((bins_raw[:-1] + 0.5 * (bins_raw[1] - bins_raw[0])) * unit - offset)
    if t_start > -np.inf or t_stop < np.inf:
        mask = (bin_centers >= t_start) * (bin_centers <= t_stop)
        bin_centers = bin_centers[mask]
        nanot_hist = nanot_hist[mask]
    if pdf:
        nanot_hist = (nanot_hist.astype('float') /
                      (nanot_hist.sum() * unit * rebin))
    return bin_centers, nanot_hist


def decay_hist_offset(nanotimes, unit, idxmax, idxmin=0, rmin=0.1, rmax=0.9,
                      cross_th=0.5, rebin=1, pdf=True):
    """Compute the offset of the decay.

    Compute the offset of the decay as the time point where the rising edge
    crosses a threshold `cross_th` as percentage its max (by default 50% of
    the max).

    Arguments:
        nanotimes (array of ints): array of raw nanotimes
        unit (float): nanotimes unit
        idxmax (int): max raw nanotime value included in the histogram.
        idxmin (int): min raw nanotime value included in the histogram.
            Default 0.
        rmin, rmax (float): fraction of the max where the monotonic
            rise starts (`vmin`) or ends (`vmax`). Valid values are in the
            (0, 1) range.
        cross_th (float): fraction of the max used as threshold to compute
            the crossing time. The returned value is the time where the
            rising edge crosses this threshold.
        rebin (int): rebinning of nanotimes histogram. Default 1 (no rebin).
        pdf (bool): if True normalize the histogram to unit area.
    """
    assert 0 < rmin < cross_th < rmax < 1
    t, hist = calc_nanotime_hist(nanotimes, unit, idxmin=idxmin,
                                 idxmax=idxmax, rebin=rebin, pdf=pdf)
    vcross = hist.max() * cross_th
    vmin, vmax = rmin * hist.max(), rmax * hist.max()

    idxpeak = hist.argmax()
    # last point below the threshold
    idxcross_minus = np.where(hist[:idxpeak] < vcross)[0][-1]
    # first point above the threshold
    idxcross_plus = np.where(hist[:idxpeak] > vcross)[0][0]

    # NOTE: This function may be simply computing t_cross
    # interpolating between t[idxcross_minus] and t[idxcross_plus].
    # However I use a larger range (defined by rmin and rmax)
    # to make sure that the rising-edge is monotonic and therefore
    # the offset computation is accurate.

    # monotonic raising edge start: last point <= vmin
    idxstart = np.where(hist[:idxpeak] <= vmin)[0][-1]
    # monotonic raising edge stop: first point >= vmax
    idxstop = np.where(hist >= vmax)[0][0]

    # The range needs to include the stop point
    ti, yi = t[idxstart:idxstop + 1], hist[idxstart:idxstop + 1]
    assert (np.diff(yi) > 0).all(), 'Raising edge non-monotonic.'
    assert idxstart <= idxcross_minus < idxcross_plus <= idxstop

    t_cross = np.interp(vcross, yi, ti)
    assert t[idxcross_minus] <= t_cross <= t[idxcross_plus]
    return t_cross



import numpy as np
import pandas as pd
from randomgen import RandomGenerator, Xoroshiro128

import pycorrelate as pyc
import depi
from depi import fret
from depi import tcspc


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Utility functions for nanotimes, FCS, etc...
#


def add_irf_nanot(burstsph_sim, irf, seed=1):
    """Add a random time to each nanotime distributed as an IRF PDF.

    Arguments:
        burstsph_sim (DataFrame): photon data, one row per photon. Columns must
            include 'nanotime' and 'stream'.
        irf (DataFrame): IRF histograms. Index is time in ns. Columns must include
            'DD_smooth1' and 'DA_smooth1'.

    Returns:
        In-place modified burstsph_sim, with new column 'nanotime_conv'.
    """
    np.random.seed(seed)
    nt_d = burstsph_sim.loc[burstsph_sim.stream == 'DexDem', 'nanotime']
    nt_a = burstsph_sim.loc[burstsph_sim.stream == 'DexAem', 'nanotime']
    IRF_t_sim = irf.index - irf.index[0]
    # smooth1 is the IRF normalized so that the sum of bin counts is 1
    # With this normalization the area of the IRF is not 1.
    _DD_valid = ~irf.DD_smooth1.isnull()
    _DA_valid = ~irf.DA_smooth1.isnull()
    IRF_DD_rnd = np.random.choice(IRF_t_sim[_DD_valid],
                                  p=irf.DD_smooth1.loc[_DD_valid],
                                  size=len(nt_d))
    IRF_DA_rnd = np.random.choice(IRF_t_sim[_DA_valid],
                                  p=irf.DA_smooth1.loc[_DA_valid],
                                  size=len(nt_a))
    nt_d_conv = nt_d + IRF_DD_rnd  # - IRF_DD_rnd.mean()
    nt_a_conv = nt_a + IRF_DA_rnd  # - IRF_DA_rnd.mean()
    assert nt_d_conv.shape[0] + nt_a_conv.shape[0] == burstsph_sim.shape[0]
    burstsph_sim['nanotime_conv'] = pd.concat([nt_d_conv, nt_a_conv]).sort_index()
    assert (burstsph_sim.loc[burstsph_sim.stream == 'DexDem', 'nanotime_conv'] == nt_d_conv).all()
    assert (burstsph_sim.loc[burstsph_sim.stream == 'DexAem', 'nanotime_conv'] == nt_a_conv).all()
    return burstsph_sim


def nanot_hist_from_burstph(burstsph, bins, col='nanotime', offset=0):
    """Return D and A nanotime histograms from burst photon-data DataFrame.
    """
    DD = burstsph.stream == 'DexDem'
    DA = burstsph.stream == 'DexAem'
    nanotimes_d = burstsph.loc[DD, col] - offset
    nanotimes_a = burstsph.loc[DA, col] - offset
    hist_params = dict(bins=bins, density=False)
    nanot_hist_d, _ = np.histogram(nanotimes_d, **hist_params)
    nanot_hist_a, _ = np.histogram(nanotimes_a, **hist_params)
    return nanot_hist_d, nanot_hist_a


def _add_bg_irf(nanot_hist_exp, nanot_hist_sim, irf, i):
    """Add background IRF generated by direct laser leakage to a nanotime histogram.

    Arguments:
        nanot_hist_exp (array): experimental nanotime histogram (i.e. fluorescence
            decay). Used to normalize the IRF peak height.
        nanot_hist_sim (array): simulated nanotime histogram (i.e. fluorescence
            decay) to which the background IRF needs to be added.
        irf (Series): IRF histogram with same binning as the nanotime histograms.
        i (int): index used to align the IRF histogram array to the nanotime histogram.

    Returns:
        Array computed summing `nanot_hist_sim` with the IRF peak. The IRF peak is
        scaled so that the simulated histograms at index `i` as the same height as
        the maximum of the experimental histogram.
    """
    irf_values = irf.loc[~irf.isnull()].values
    irf_values /= irf_values.max()
    i_irf_max = irf_values.argmax()
    delta_irf = nanot_hist_exp.max() - nanot_hist_sim[i_irf_max + i]
    nanot_hist_sim_pirf = nanot_hist_sim.copy()
    nanot_hist_sim_pirf[i:i + irf_values.shape[0]] = (
        nanot_hist_sim[i:i + irf_values.shape[0]]
        + delta_irf * irf_values
    )
    return nanot_hist_sim_pirf


def calc_nanot_hist_irf_da(bph_sim, irf, nt_bins, tcspc_unit,
                           nanot_hist_d_exp, nanot_hist_a_exp,
                           add_bg_irf=False, irf_seed=1):
    """Compute simulated D and A fluorescence decays histograms with IRF.
    The IRF is added to the nanotimes before computing the histogram. Optionally,
    when `add_bg_irf=True`, an IRF peak is also added to the nanotime histogram
    to simulate direct laser leakage. See :func:`_add_bg_irf` for details.
    The returned histograms are normalized to the same area as the experimental
    decays passed as argument.
    """
    # Add IRF to nanotimes
    bph_sim = add_irf_nanot(bph_sim, irf, seed=irf_seed)  # adds column `nanotime_conv`
    # Convert nanotimes in raw units as the experiment
    bph_sim['nanotime_conv_unit'] = (
        (bph_sim.nanotime_conv * (1e-9 / tcspc_unit)).round().astype('int'))
    # Find the offset, the raise-time of the decay
    offset_Dex_sim = tcspc.decay_hist_offset(bph_sim.nanotime_conv_unit,
                                             tcspc_unit, 4095, rebin=4) * 1e9
    # Compute DexDem and DexAem nanotime histograms
    nanot_hist_d_sim, nanot_hist_a_sim = nanot_hist_from_burstph(
        bph_sim, bins=nt_bins, col='nanotime_conv_unit',
        offset=offset_Dex_sim / (1e9 * tcspc_unit)
    )
    if add_bg_irf:
        nanot_hist_d_sim = _add_bg_irf(nanot_hist_d_exp, nanot_hist_d_sim, irf.DD_smooth1, i=129)
        nanot_hist_a_sim = _add_bg_irf(nanot_hist_a_exp, nanot_hist_a_sim, irf.DA_smooth1, i=125)
    # Normalize to the same area as experimental decays
    nanot_hist_d_sim = nanot_hist_d_sim * nanot_hist_d_exp.sum() / nanot_hist_d_sim.sum()
    nanot_hist_a_sim = nanot_hist_a_sim * nanot_hist_a_exp.sum() / nanot_hist_a_sim.sum()
    return nanot_hist_d_sim, nanot_hist_a_sim


def calc_fcs_dd_da(burstsph, bins, seed=1):
    ts_d = burstsph.loc[burstsph.stream == 'DexDem', 'timestamp'].values.copy()
    ts_a = burstsph.loc[burstsph.stream == 'DexAem', 'timestamp'].values.copy()
    np.random.seed(seed)
    ts_d += np.random.randint(-25, 25, size=ts_d.size)
    ts_a += np.random.randint(-25, 25, size=ts_a.size)
    CC_DA = pyc.pcorrelate(ts_d, ts_a, bins, normalize=True) + 1
    AC_DD = pyc.pcorrelate(ts_d, ts_d, bins, normalize=True) + 1
    return CC_DA, AC_DD


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Metrics functions
#


def mse_metric(y_sim, y_data, weights=None):
    """Mean Square Error statistics"""
    if weights is None:
        weights = np.ones_like(y_sim)
    return (((y_sim - y_data) / weights)**2).mean()


def loglike_metric(y_sim, y_data):
    """Log-likelihood statistics for Poisson data (example: bin counts)
    """
    v = y_sim > 0
    return (y_sim[v] - y_data[v] * np.log(y_sim[v])).sum()


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Loss functions starting with `iloss` take the DataFrame of simulated photons as input
#


def iloss_loglike_E(bph_sim, E_exp, E_bins=np.arange(0, 1.1, 0.03)):
    """Log-likelihood of simulated E histograms vs experimental
    """
    E = fret.calc_E_burst(bph_sim)
    E_hist_sim, _ = np.histogram(E, bins=E_bins)
    E_hist_exp, _ = np.histogram(E_exp, bins=E_bins)
    ll = loglike_metric(E_hist_sim, E_hist_exp)
    ll0 = loglike_metric(E_hist_exp, E_hist_exp)
    return ll - ll0


def iloss_loglike_nanot(bph_sim, irf, nt_bins, tcspc_unit,
                        nanot_hist_d_exp, nanot_hist_a_exp,
                        loglike_d_std, loglike_a_std, add_bg_irf=False,
                        irf_seed=1, return_da_losses=False):
    """Log-likelihood of simulated nantimes histograms vs experimental
    """
    nanot_hist_d_sim, nanot_hist_a_sim = calc_nanot_hist_irf_da(
        bph_sim, irf, nt_bins, tcspc_unit, nanot_hist_d_exp, nanot_hist_a_exp,
        add_bg_irf=add_bg_irf, irf_seed=irf_seed)
    # Compute log-likelihood metrics
    ll0_d = loglike_metric(nanot_hist_d_exp, nanot_hist_d_exp)
    ll0_a = loglike_metric(nanot_hist_a_exp, nanot_hist_a_exp)
    loglike_d = loglike_metric(nanot_hist_d_sim, nanot_hist_d_exp) - ll0_d
    loglike_a = loglike_metric(nanot_hist_a_sim, nanot_hist_a_exp) - ll0_a
    loglike_tot = loglike_d / loglike_d_std + loglike_a / loglike_a_std
    if return_da_losses:
        return loglike_tot, loglike_d, loglike_a
    else:
        return loglike_tot


def iloss_residuals_fcs(bph_sim, bins, CC_DA_exp, AC_DD_exp,
                        CC_DA_std_dev, AC_DD_std_dev,
                        CC_DA_loss_std=1, AC_DD_loss_std=1,
                        tot_loss_std=1):
    """Residuals of FCS curves vs experimental
    """
    CC_DA_sim, AC_DD_sim = calc_fcs_dd_da(bph_sim, bins)
    loss_cc_da = mse_metric(CC_DA_sim, CC_DA_exp, weights=CC_DA_std_dev)
    loss_ac_dd = mse_metric(AC_DD_sim, AC_DD_exp, weights=AC_DD_std_dev)
    loss_fcs = (loss_cc_da / CC_DA_loss_std + loss_ac_dd / AC_DD_loss_std)
    return loss_fcs


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Functions starting with `loss` take the simulation parameters as input arguments.
# The first 4 positional arguments are always: params_vary, params_const, ts, recolor_func
# `recolor_func` is a specific "recoloring function" (see next section) which determines
# the D-A distance distribution and which arguments are varying.
#
# The keyword argument `cache` is passed to the recoloring function.
#

def loss_E(params_vary, params_const, ts, recolor_func, E_exp, E_bins,
           cache=False, seed=1):
    """Loss function for E histogram.
    """
    bph_sim = recolor_func(params_vary, params_const, ts, cache=cache, seed=seed)
    loss_E = iloss_loglike_E(bph_sim, E_exp=E_exp, E_bins=E_bins)
    return loss_E


def loss_nanot(params_vary, params_const, ts, recolor_func,
               irf, nt_bins, tcspc_unit, nanot_hist_d_exp, nanot_hist_a_exp,
               loglike_d_std, loglike_a_std, add_bg_irf=False, irf_seed=1,
               cache=False, seed=1):
    """Loss function for fluorescence decays.
    """
    bph_sim = recolor_func(params_vary, params_const, ts, cache=cache, seed=seed)
    loss_nanot = iloss_loglike_nanot(
        bph_sim, irf=irf, nt_bins=nt_bins, tcspc_unit=tcspc_unit,
        nanot_hist_d_exp=nanot_hist_d_exp, nanot_hist_a_exp=nanot_hist_a_exp,
        loglike_d_std=loglike_d_std, loglike_a_std=loglike_a_std,
        add_bg_irf=add_bg_irf, irf_seed=irf_seed)
    return loss_nanot


def loss_E_fcs(params_vary, params_const, ts, recolor_func, E_exp, E_bins,
               fcs_bins, CC_DA_exp, AC_DD_exp, CC_DA_std_dev, AC_DD_std_dev,
               CC_DA_loss_std, AC_DD_loss_std,
               E_loss_std=1, FCS_loss_std=1,
               cache=False, seed=1):
    """Loss function combining E histogram and FCS curves (CC_DA and AC_DD).
    """
    bph_sim = recolor_func(params_vary, params_const, ts, cache=cache, seed=seed)
    loss_E = iloss_loglike_E(bph_sim, E_exp=E_exp, E_bins=E_bins)
    loss_fcs = iloss_residuals_fcs(
        bph_sim, bins=fcs_bins, CC_DA_exp=CC_DA_exp, AC_DD_exp=AC_DD_exp,
        CC_DA_std_dev=CC_DA_std_dev, AC_DD_std_dev=AC_DD_std_dev,
        CC_DA_loss_std=CC_DA_loss_std, AC_DD_loss_std=AC_DD_loss_std)
    return (loss_E / E_loss_std + loss_fcs / FCS_loss_std)


def loss_E_nanot(params_vary, params_const, ts, recolor_func, E_exp, E_bins,
                 irf, nt_bins, tcspc_unit, nanot_hist_d_exp, nanot_hist_a_exp,
                 loglike_d_std, loglike_a_std, add_bg_irf=False, irf_seed=1,
                 E_loss_std=1, nanot_loss_std=1,
                 cache=False, seed=1):
    """Loss function combining E histogram and fluorescence decays.
    """
    bph_sim = recolor_func(params_vary, params_const, ts, cache=cache, seed=seed)
    loss_E = iloss_loglike_E(bph_sim, E_exp=E_exp, E_bins=E_bins)
    loss_nanot = iloss_loglike_nanot(
        bph_sim, irf=irf, nt_bins=nt_bins, tcspc_unit=tcspc_unit,
        nanot_hist_d_exp=nanot_hist_d_exp, nanot_hist_a_exp=nanot_hist_a_exp,
        loglike_d_std=loglike_d_std, loglike_a_std=loglike_a_std,
        add_bg_irf=add_bg_irf, irf_seed=irf_seed)
    return (loss_E / E_loss_std + loss_nanot / nanot_loss_std)


def loss_E_fcs_nanot(params_vary, params_const, ts, recolor_func, E_exp, E_bins,
                     fcs_bins, CC_DA_exp, AC_DD_exp, CC_DA_std_dev, AC_DD_std_dev,
                     CC_DA_loss_std, AC_DD_loss_std,
                     irf, nt_bins, tcspc_unit, nanot_hist_d_exp, nanot_hist_a_exp,
                     loglike_d_std, loglike_a_std, add_bg_irf=False, irf_seed=1,
                     E_loss_std=1, FCS_loss_std=1, nanot_loss_std=1,
                     cache=False, seed=1):
    """Loss function combining E histogram, FCS and fluorescence decays.
    """
    bph_sim = recolor_func(params_vary, params_const, ts, cache=cache, seed=seed)
    loss_E = iloss_loglike_E(bph_sim, E_exp=E_exp, E_bins=E_bins)
    loss_fcs = iloss_residuals_fcs(
        bph_sim, bins=fcs_bins, CC_DA_exp=CC_DA_exp, AC_DD_exp=AC_DD_exp,
        CC_DA_std_dev=CC_DA_std_dev, AC_DD_std_dev=AC_DD_std_dev,
        CC_DA_loss_std=CC_DA_loss_std, AC_DD_loss_std=AC_DD_loss_std)
    loss_nanot = iloss_loglike_nanot(
        bph_sim, irf=irf, nt_bins=nt_bins, tcspc_unit=tcspc_unit,
        nanot_hist_d_exp=nanot_hist_d_exp, nanot_hist_a_exp=nanot_hist_a_exp,
        loglike_d_std=loglike_d_std, loglike_a_std=loglike_a_std,
        add_bg_irf=add_bg_irf, irf_seed=irf_seed)
    return (loss_E / E_loss_std + loss_fcs / FCS_loss_std + loss_nanot / nanot_loss_std)


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# The functions `recolor_burstsph_*` (or "recoloring functions") call the recoloring
# simulation for different distance distribution and different sets of varying
# parameters.
# The first 4 arguments are always: params, params_const, ts, cache
#


def _recolor_burstsph(ts, params, cache=False, seed=1):
    if cache:
        bph = depi.recolor_burstsph_cache(ts, **params)
    else:
        rg = RandomGenerator(Xoroshiro128(seed))
        bph = depi.recolor_burstsph(ts, rg=rg, **params)
    return bph


def recolor_burstsph_gauss_noblink(params, params_const, ts, cache=True, seed=1):
    R_mean, R_sigma, τ_relax = params
    params_tot = params_const.copy()
    params_tot.update(R_mean=R_mean, R_sigma=R_sigma, τ_relax=τ_relax)
    return _recolor_burstsph(ts, params_tot, cache=cache, seed=1)


def recolor_burstsph_gauss(params, params_const, ts, cache=True, seed=1):
    R_mean, R_sigma, τ_relax, prob_A_dark, τ_A_dark = params
    τ_A_dark *= 1e6
    params_tot = params_const.copy()
    params_tot.update(R_mean=R_mean, R_sigma=R_sigma,
                      τ_relax=τ_relax, prob_A_dark=prob_A_dark,
                      τ_A_dark=τ_A_dark)
    return _recolor_burstsph(ts, params_tot, cache=cache, seed=1)


def recolor_burstsph_wlc(params, params_const, ts, cache=True, seed=1):
    L, lp, τ_relax, prob_A_dark, τ_A_dark = params
    τ_A_dark *= 1e6
    params_tot = params_const.copy()
    params_tot.update(L=L, lp=lp,
                      τ_relax=τ_relax, prob_A_dark=prob_A_dark,
                      τ_A_dark=τ_A_dark)
    return _recolor_burstsph(ts, params_tot, cache=cache, seed=1)


def recolor_burstsph_wlco(params, params_const, ts, cache=True, seed=1):
    L, lp, offset, τ_relax, prob_A_dark, τ_A_dark = params
    τ_A_dark *= 1e6
    params_tot = params_const.copy()
    params_tot.update(L=L, lp=lp, offset=offset,
                      τ_relax=τ_relax, prob_A_dark=prob_A_dark,
                      τ_A_dark=τ_A_dark)
    return _recolor_burstsph(ts, params_tot, cache=cache, seed=1)


def recolor_burstsph_gauss_nodiff(params, params_const, ts, cache=True, seed=1):

    R_mean, R_sigma, prob_A_dark, τ_A_dark = params
    τ_A_dark *= 1e6
    params_tot = params_const.copy()
    params_tot.update(R_mean=R_mean, R_sigma=R_sigma,
                      prob_A_dark=prob_A_dark,
                      τ_A_dark=τ_A_dark)
    return _recolor_burstsph(ts, params_tot, cache=cache, seed=1)


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# These function are here for compatibility with old notebooks.
# Do not use in new code. Eventually they will be deleted.
# All the functions below can be replace by the function `loss_E`
# which takes a recoloring function. The recoloring function determines
# the distance distribution and which arguments are varying.
#

def loss_function_E(params, params_const, ts, E_exp,
                    bins=np.arange(0, 1.1, 0.03), cache=True):
    """Loss function for E histograms and Gaussian distance distribution
    """
    bph = recolor_burstsph_gauss(params, params_const, ts, cache=cache)
    return iloss_loglike_E(bph, E_exp, E_bins=bins)


def loss_function_E_wlc(params, params_const, ts, E_exp,
                        bins=np.arange(0, 1.1, 0.03), cache=True):
    """Loss function for E histograms and WLC distance distribution (no offset)
    """
    bph = recolor_burstsph_wlc(params, params_const, ts, cache=cache)
    return iloss_loglike_E(bph, E_exp, E_bins=bins)


def loss_function_E_wlco(params, params_const, ts, E_exp,
                         bins=np.arange(0, 1.1, 0.03), cache=True):
    """Loss function for E histograms and WLC distance distribution with offset
    """
    bph = recolor_burstsph_wlco(params, params_const, ts, cache=cache)
    return iloss_loglike_E(bph, E_exp, E_bins=bins)


def loss_function_E_gauss_nodiff(params, params_const, ts, E_exp,
                                 bins=np.arange(0, 1.1, 0.03), cache=True):
    """Loss function for E histograms and Gaussian distance (fixed τ_relax)
    """
    bph = recolor_burstsph_gauss_nodiff(params, params_const, ts, cache=cache)
    return iloss_loglike_E(bph, E_exp, E_bins=bins)



"""
Compute BVA from burst photon data.

The input photon data must be a DataFrame with one row per photon and
mandatory columns 'timestamp' and 'stream' (additional columns are ignored).
The input photons must be only the photons remaining after burst search.

The DataFrame index must contains two levels ('burst', 'ph'). 'burst' is
an integer indicating the burst number. 'ph' is an integer indicating the
photon-number in each burst ('ph' is 0 for the first photon of each burst).
This format is exported by FRETBursts when saving the burst photon data.
"""

from collections import defaultdict
import numpy as np
import pandas as pd


def bva(burstsph_Dex, n):
    """Compute BVA quantity for each burst.

    Arguments:
        burstsph_Dex (pandas.DataFrame): dataframe with one row per photon
            and a column `stream` indicating the stream of each photon.
            The index need to have two levels (burst, ph) indicating the
            burst and photon ID respectively.
        n (int): the number of photon used to compute E in sub-bursts
    """
    E_sub_std = []
    DexAem = burstsph_Dex.stream == 'DexAem'
    for i in np.unique(burstsph_Dex.index.get_level_values('burst')):
        DexAem_burst = DexAem.loc[i].values
        E_sub_bursts = []
        for istart in range(0, DexAem_burst.shape[0], n):
            A_D = DexAem_burst[istart:istart + n].sum()
            E = A_D / n
            E_sub_bursts.append(E)
        E_sub_std.append(np.std(E_sub_bursts))
    E_sub_std = np.array(E_sub_std)
    return E_sub_std


def bva_bin(burstsph_Dex, n, num_sub_bursts_th=60,
            E_bins=np.arange(0, 1.01, 0.05)):
    """Compute the mean BVA on E bins.

    Arguments:
        burstsph_Dex (pandas.DataFrame): dataframe with one row per photon
            and a column `stream` indicating the stream of each photon.
            The index need to have two levels (burst, ph) indicating the
            burst and photon ID respectively.
        n (int): the number of photon used to compute E in sub-bursts
        num_sub_bursts_th (int): minimum number of sub-bursts in a bin
            for BVA to be reported.
        E_bins (array): array of E bin edges, size in number of bins + 1.

    Returns:
        2-tuple of two arrays:
        - `E_sub_std_mean_bin`: the mean BVA in each E bins
        - `E_centers`: bin centers for each value in `E_sub_std_mean_bin`.
    """
    E_centers = E_bins[:-1] + 0.5 * (E_bins[1] - E_bins[0])
    burstsph_Dex_grp = burstsph_Dex.groupby('burst')
    # compute raw E values from photon data
    Nd_plus_Na = burstsph_Dex_grp['stream'].count()
    Na = burstsph_Dex_grp['stream'].agg(lambda x: (x == 'DexAem').sum())
    E_raw = Na / Nd_plus_Na
    # index of the bin for each burst in E_raw
    burst_indices = np.unique(burstsph_Dex.index.get_level_values('burst'))
    E_raw_bins = pd.Series(np.digitize(E_raw, E_bins) - 1,
                           index=burst_indices)
    E_sub_bursts_bin = defaultdict(list)
    DexAem = burstsph_Dex.stream == 'DexAem'
    for i in np.unique(burstsph_Dex.index.get_level_values('burst')):
        DexAem_burst = DexAem.loc[i].values
        E_raw_bin = E_raw_bins.loc[i]
        for istart in range(0, DexAem_burst.shape[0], n):
            A_D = DexAem_burst[istart:istart + n].sum()
            E = A_D / n
            E_sub_bursts_bin[E_raw_bin].append(E)
    E_sub_std_mean_bin = \
        [np.std(E_sub_bursts_bin[i])
         if len(E_sub_bursts_bin[i]) > num_sub_bursts_th else -1
         for i in range(len(E_centers))]
    assert len(E_sub_std_mean_bin) == len(E_bins) - 1
    return np.array(E_sub_std_mean_bin), E_centers


def bva_bin_combo(burstsph_Dex, n, num_sub_bursts_th=60,
                  E_bins=np.arange(0, 1.01, 0.05)):
    """Compute the BVA per each burst and mean BVA on E bins.

    This function combines :func:`bva` and :func:`bva_bin` and is much
    faster than calling these two functions separately.

    Arguments:
        burstsph_Dex (pandas.DataFrame): dataframe with one row per photon
            and a column `stream` indicating the stream of each photon.
            The index need to have two levels (burst, ph) indicating the
            burst and photon ID respectively.
        n (int): the number of photon used to compute E in sub-bursts
        num_sub_bursts_th (int): minimum number of sub-bursts in a bin
            for BVA to be reported.
        E_bins (array): array of E bin edges, size in number of bins + 1.

    Returns:
        Tuple of 3 arrays:
        - `E_sub_std`: the BVA computed on each burst
        - `E_sub_std_mean_bin`: the mean BVA in each E bins
        - `E_centers`: bin centers for each value in `E_sub_std_mean_bin`.
    """
    E_centers = E_bins[:-1] + 0.5 * (E_bins[1] - E_bins[0])
    burstsph_Dex_grp = burstsph_Dex.groupby('burst')
    # compute raw E values from photon data
    E_raw = (burstsph_Dex_grp['stream']
             .agg(lambda x: (x == 'DexAem').sum() / x.size))
    # index of the left bin edges in E_bins for each burst in E_raw
    burst_indices = np.unique(burstsph_Dex.index.get_level_values('burst'))
    E_raw_bins = pd.Series(np.digitize(E_raw, E_bins) - 1,
                           index=burst_indices)
    # standard deviation of sub-bursts in each burst
    E_sub_std = []
    # dict with bin index as key and values that are a list of lists.
    # The innermost list contains the E values of sub-bursts in a burst.
    E_sub_bursts_bin = defaultdict(list)
    DexAem = burstsph_Dex.stream == 'DexAem'
    # Loop over each burst
    for i in np.unique(burstsph_Dex.index.get_level_values('burst')):
        DexAem_burst = DexAem.loc[i].values
        E_raw_bin = E_raw_bins.loc[i]
        # We initialize a list of sub-bursts for each burst
        E_sub_bursts = []
        # Loop over each n-tuple of photons (i.e. sub-bursts)
        for istart in range(0, DexAem_burst.shape[0], n):
            A_D = DexAem_burst[istart:istart + n].sum()
            E = A_D / n
            E_sub_bursts.append(E)
        # Append list of current burst as an element in the list of current bin
        E_sub_bursts_bin[E_raw_bin].append(E_sub_bursts)
        E_sub_std.append(np.std(E_sub_bursts))
    E_bin_len = {k: sum([len(x) for x in v])
                 for k, v in E_sub_bursts_bin.items()}
    E_sub_std_mean_bin = [np.std(np.hstack(E_sub_bursts_bin[i]))
                          if E_bin_len.get(i, 0) > num_sub_bursts_th else -1
                          for i in range(len(E_centers))]
    E_sub_std = np.array(E_sub_std)
    assert len(E_sub_std_mean_bin) == len(E_centers)
    return E_sub_std, np.array(E_sub_std_mean_bin), E_centers
